export class BaseEnemy {
    constructor(instance, player, runtime, level) {
        this.instance = instance; this.player = player; this.runtime = runtime;
        this.level = level; this.speed = 50; this.hp = 50 + (5 * this.level);  this.maxHP = this.hp; this.damage = 15;
        this.levelUI = this.runtime.objects.Level.createInstance(1, this.instance.x, this.instance.y);
        this.hpBar = this.runtime.objects.EnemyHealthBar.createInstance(0, this.instance.x, this.instance.y);
        this.knockbackVelocity = { x: 0, y: 0}; this.knockbackDecay = 0.9; this.minKnockbackSpeed = 1; this.invulnerableTime = 0;
        this.isProtected = false; this.shield = null; this.shieldTimer = 0;
    }
    update() {
        if (this.instance == null || this.player == null) return;
        this.updateKnockback();
        this.levelUI.setAnimation(String(this.level)); this.levelUI.x = this.instance.x; this.levelUI.y = this.instance.y;
        if (this.instance.testOverlap(Game.slashHitbox)) {	// Detects collsion bet. enemy and sword
            if(Game.player.isAttacking) {
                this.takeDamage(Game.vars.playerDamage); Game.player.applyKnockback(this.instance.x, Game.player.instance.y, Game.vars.playerRecoil);
			}
        }
        let swordBeams = this.runtime.objects.SwordBeam.getAllInstances();
        for (const beam of swordBeams) {
            if (this.instance.testOverlap(beam)) {	// Detects collsion bet. enemy and sword beam
                this.takeDamage(Game.vars.playerDamage/4);
                beam.destroy();
        }}

        if (this.invulnerableTime > 0) {this.invulnerableTime -= this.runtime.dt;}  // I-frames
        else {this.instance.setAnimation("Idle");}
        
        this.hpBar.width = 25 * (this.hp / this.maxHP); this.hpBar.x = this.instance.x; this.hpBar.y = this.instance.y;//hp bar
        if (this.hp < this.maxHP && this.hp > 0) {this.hpBar.isVisible = true;} else {this.hpBar.isVisible = false;}

        if (this.player.x > this.instance.x) {this.instance.width = 24;}
        else {this.instance.width = -24}
        if (this.hp <= 0) {this.terminate();}

        if (this.instance == null || this.player == null) return;
        if (this.shieldTimer > 0) {this.shield.moveToTop(); this.shield.x = this.instance.x; this.shield.y = this.instance.y;}
        else {if (this.shield != null) {this.shield.destroy()};}
    }
    takeDamage(damage) {
        if (this.invulnerableTime > 0) {return;} 
        this.applyKnockback(this.player.x, this.player.y, Game.vars.playerKnockback);
        this.runtime.globalVars.ScreenShake = true;
        const damageSparks = this.runtime.objects.DamageSparks.createInstance(1, this.instance.x, this.instance.y);
        if (!this.isProtected) {
            this.runtime.callFunction("playAudio", "EnemyHit", -5, false, "");
            this.hp -= damage; this.invulnerableTime = .25; this.instance.setAnimation("Flash"); console.log("enemy was not protected");
        } else {
            this.shield = this.runtime.objects.ProtectionCrystal.createInstance(1, this.instance.x, this.instance.y);
            this.shieldTimer = .25; this.runtime.callFunction("playAudio", "Negate", -10, false, ""); console.log("enemy was protected");
        } 
    }
    terminate() {
        const deathEffect = this.runtime.objects.EnemyDamageParticles.createInstance(1, this.instance.x, this.instance.y);
        let hpSpawnChance = randInt(0, 100);
        if (hpSpawnChance >= 0 && hpSpawnChance <= 25 + this.runtime.globalVars.room) {let pickup = this.runtime.objects.HPpickup.createInstance(1, this.instance.x, this.instance.y);}
        this.instance.destroy(); this.instance = null; this.levelUI.destroy(); this.runtime.callFunction("playAudio", "CrystalDeath", -10, false, "");
    }
    applyKnockback(sourceX, sourceY, magnitude) {
        const dx = this.instance.x - sourceX;  // Get direction from damage source
        const dy = this.instance.y - sourceY;
        const dist = Math.sqrt(dx * dx + dy * dy);
        if (dist === 0) {return};
        
        const nx = dx / dist;   // Normalize and apply knockback
	    const ny = dy / dist;
	    this.knockbackVelocity.x += nx * magnitude;
	    this.knockbackVelocity.y += ny * magnitude;
    }
    updateKnockback() {
		const v = this.knockbackVelocity; const speed = Math.sqrt(v.x * v.x + v.y * v.y);
		if (speed > this.minKnockbackSpeed) {
			this.instance.x += v.x * this.runtime.dt; this.instance.y += v.y * this.runtime.dt;
            v.x *= this.knockbackDecay; v.y *= this.knockbackDecay;
		} else {v.x = 0; v.y = 0;}
	}
    followPlayer() {
        if (this.instance == null || this.player == null) return;
        const dx = this.player.x - this.instance.x; const dy = this.player.y - this.instance.y;
        const angle = Math.atan2(dy, dx);
        this.instance.x += Math.cos(angle) * this.speed * this.runtime.dt;
        this.instance.y += Math.sin(angle) * this.speed * this.runtime.dt;
    }
}
export class LaserEnemy extends BaseEnemy {
    constructor(instance, player, runtime, level) {
        super(instance, player, runtime, level);
        this.speed = 20 + this.runtime.globalVars.speedBoost; this.hp = 30 + (5 * this.level); this.maxHP = this.hp; this.damage = 5 + (2 * this.level);
        this.laser = null; this.laserTimer = 0; this.chargingLaser = false; this.laserRecoil = 200;
        setInterval(() => this.chargeLaser(), 4000 - this.runtime.globalVars.fireSpeed);
    }
    update() {
        super.update();
        if (this.instance == null || this.player == null) {this.runtime.callFunction("stopAudio", "Beep"); return;}
        super.followPlayer();
        if (this.laserTimer > 0) {
            this.laserTimer -= this.runtime.dt;
            this.laser.x = this.instance.x; this.laser.y = this.instance.y;
        }
        if (this.laserTimer <= 0 && this.chargingLaser) {
            this.chargingLaser = false;
            let recoilSourceX = this.laser.x + Math.cos(this.laser.angle) * 100;
            let recoilSourceY = this.laser.y + Math.sin(this.laser.angle) * 100;
            super.applyKnockback(recoilSourceX, recoilSourceY, this.laserRecoil);
            this.runtime.callFunction("playAudio", "LaserReleased", -10, false, ""); this.runtime.callFunction("stopAudio", "Beep");
        }
    }
    chargeLaser() {
        if (this.instance == null || this.player == null) return;
        this.chargingLaser = true; this.laserTimer = 2;
        this.runtime.callFunction("playAudio", "Beep", -20, true, "slow");
        this.laser = this.runtime.objects.Laser.createInstance(1, this.instance.x, this.instance.y); 
    }
    terminate() {
        super.terminate(); if (this.laser == null) return; this.laser.destroy();
    }
}
export class RushEnemy extends BaseEnemy {
    constructor(instance, player, runtime, level) {
        super(instance, player, runtime, level)
        this.speed = 35 + this.runtime.globalVars.speedBoost; this.hp = 25 + (5 * this.level); this.maxHP = this.hp; this.damage = 5 + (1.5 * this.level);
        this.slowTimer = 0; this.justFired = false;
        setInterval(() => this.fire(), 3500 - this.runtime.globalVars.fireSpeed);
    }
    update() {
        super.update();
        super.followPlayer();
        if (this.slowTimer > 0) {
            this.slowTimer -= this.runtime.dt;
            this.speed = 20;
            if (this.invulnerableTime <= 0) {this.instance.setAnimation("Stop");}
            if (!this.justFired && this.slowTimer < 0.2) {
                this.justFired = true;
                this.runtime.callFunction("playAudio", "Fire", -15, false, "");
                for (let i = 0; i < 8; i++) {
                    let bullet = this.runtime.objects.Bullet.createInstance(1, this.instance.x, this.instance.y);
                    bullet.angle = i * 45 * (Math.PI/180);
                }
            }
        } else {this.speed = 35; this.justFired = false;}
    }
    fire() {if (this.instance == null) return; this.runtime.callFunction("playAudio", "HurtSound", -10, false, ""); this.slowTimer = 2;}
}
export class JumpEnemy extends BaseEnemy {
    constructor(instance, player, runtime, level) {
        super(instance, player, runtime, level)
        this.speed = 30 + this.runtime.globalVars.speedBoost; this.hp = 40 + (5 * this.level); this.maxHP = this.hp; this.damage = 5 + (2 * this.level);
        this.platform = this.instance.behaviors.Platform; this.direction = "";
        setInterval(() => this.jump(), 2000 - this.runtime.globalVars.fireSpeed);
    }
    update() {
        super.update();
        if (this.platform.isJumping || this.platform.isFalling) {
            if (this.invulnerableTime <= 0) {this.instance.setAnimation("jump");} 
            if (this.direction == "right") {this.platform.simulateControl("right");}
            else if (this.direction == "left") {this.platform.simulateControl("left");}
        } 
    }
    jump() {
        if (this.instance == null || this.player == null) return;
        this.platform.simulateControl("jump"); this.runtime.callFunction("playAudio", "FrogJump", -10, false, "");
        if (this.player.x > this.instance.x) {this.direction = "right";}
        else {this.direction = "left";}
    }
}
export class WalkEnemy extends BaseEnemy {
    constructor(instance, player, runtime, level) {
        super(instance, player, runtime, level)
        this.speed = 20 + this.runtime.globalVars.speedBoost; this.hp = 40 + (5 * this.level); this.maxHP = this.hp; this.damage = 5 + (1.5 * this.level);
        this.platform = this.instance.behaviors.Platform; this.slowTimer = 0; this.justFired = false;
        setInterval(() => this.fire(), 3000 - this.runtime.globalVars.fireSpeed);
    }
    update() {
        super.update();
        if (this.instance == null || this.player == null) return;
        this.platform.maxSpeed = this.speed;
        let sign = (this.instance.width/24);
        if (this.slowTimer > 0) {
            this.slowTimer -= this.runtime.dt;
            this.speed = 0;
            if (this.invulnerableTime <= 0) {this.instance.setAnimation("Open");}
            if (!this.justFired && this.slowTimer < 0.2) {
                this.justFired = true; this.runtime.callFunction("playAudio", "Fire", -15, false, "");
                for (let i = 0; i < 6; i++) {
                    let bullet = this.runtime.objects.Bullet.createInstance(1, this.instance.x, this.instance.y);
                    bullet.angle = (Math.PI/180) * (270 + (sign * (10 * i)));
                }
            }
        } else {this.speed = 15; this.justFired = false;}
        if (this.player.x > this.instance.x) {this.platform.simulateControl("right");}
        else {this.platform.simulateControl("left");}
    }
    fire() {if (this.instance == null) return; this.runtime.callFunction("playAudio", "HurtSound", -10, false, ""); this.slowTimer = 1;}
}
export class ProtectionEnemy extends BaseEnemy {
    constructor(instance, player, runtime, level) {
        super(instance, player, runtime, level);
        this.speed = 20; this.hp = 60 + (10 * this.level); this.maxHP = this.hp; this.damage = 10;
        this.rays = [];
        for (const enemy of Game.enemies){
            if (enemy.instance == this.instance) continue;
            const ray = this.runtime.objects.ProtectiveRay.createInstance(1, this.instance.x, this.instance.y);
            const dx = enemy.instance.x - this.instance.x; const dy = enemy.instance.y - this.instance.y;
            ray.angle = Math.atan2(dy, dx); ray.width = Math.sqrt(dx * dx + dy * dy);
            this.rays.push({ ray, target: enemy }); console.log(this.rays);
        }
    }
    update() {
        super.update();
        if (this.instance == null || this.player == null) return;
        this.instance.width = 24;
        for (const { ray, target } of this.rays) {
            if (target.instance == null) continue;
            const dx = target.instance.x - this.instance.x; const dy = target.instance.y - this.instance.y;
            ray.x = this.instance.x; ray.y = this.instance.y; ray.angle = Math.atan2(dy, dx); ray.width = Math.sqrt(dx * dx + dy * dy);
            target.isProtected = true;
        }
    }
    applyKnockback() {}
    terminate() {
        for (const enemy of Game.enemies) {enemy.isProtected = false;}
        super.terminate();
        for (const { ray, target } of this.rays) {ray.destroy();}
    }
}
export class FairyEnemy extends BaseEnemy {
    constructor(instance, player, runtime, level) {
        super(instance, player, runtime, level);
        this.speed = 25 + this.runtime.globalVars.speedBoost; this.hp = 40 + (5 * this.level); this.maxHP = this.hp; this.damage = 5 + (2 * this.level);
    }
    update(){
    super.update();
    if (this.instance == null || this.player == null) return;
    super.followPlayer();
    }
    takeDamage(damage) {
        super.takeDamage(damage);
        this.instance.x = randInt(0, 320); this.instance.y = randInt(0, 180);    
    }
}
export class BossEnemy extends BaseEnemy{
    constructor(instance, player, runtime, level) {
        super(instance, player, runtime, level);
        this.hp = 50 + (10 * this.level); this.maxHP = this.hp; this.damage = 15 + this.runtime.globalVars.lvlBoost; this.timesHit = 0;
    }
    takeDamage(damage) {
        if (this.invulnerableTime > 0) {return;}
        this.timesHit += 1;
        if (this.timesHit == 3) {this.instance.instVars.attack = randInt(1, 4); this.timesHit = 0;}
        super.takeDamage(damage);
    }
}
function randInt(min, max) {return Math.floor(Math.random() * (max - min + 1)) + min;}