import { PlayerClass } from "./player.js";
import { BaseEnemy, LaserEnemy, RushEnemy, JumpEnemy, WalkEnemy, ProtectionEnemy, FairyEnemy, BossEnemy} from "./enemy.js";

globalThis.Game = {
	player: null,
	enemies: [], spikes: [],
	camera: null, slashHitbox: null, damageFlash: null, hpPickups: null,
	vars: {
		playerKnockback: 100, playerDamage: 10, playerHp: 30, playerMaxHp: 30, playerDef: 5, dashStrength: 300, dashCooldown: 1.5, maxJumps: 1, iframeDash: false, redSword: false, playerRecoil: 50, playerInvulnerabilty: 2, jumpStrength: 275, holySword: false, alive: true, fallSpeed: 175,
		wave: 0,
		nextWave: false, waveCountdown: 0,
		projectileDmg: 5,
	},
	reset() {
		this.player = null;
		this.enemies = []; this.spikes = [];
		this.camera = null; this.slashHitbox = null; this.damageFlash = null;
		this.wave = 0;
	}
};
function initializeGameObjects(runtime) {
	Game.reset(); console.log("game reset");

	globalThis.playerInstance = runtime.objects.Player.getFirstInstance();
	if (!playerInstance) {console.warn("No player instance found."); return;}

	Game.player = new PlayerClass(playerInstance, Game.vars.playerHp, runtime);
	Game.camera = runtime.objects.Camera.createInstance(0, 160, 90);
	Game.slashHitbox = runtime.objects.SlashHitBox.createInstance(0, playerInstance.x, playerInstance.y);
	Game.damageFlash = runtime.objects.PlayerFlash.createInstance(1, playerInstance.x, playerInstance.y);
	const boss = runtime.objects.Boss.getFirstInstance();
	if (boss != null) {Game.enemies.push(new BossEnemy(boss, playerInstance, runtime, 20));}
}
function endLayout(runtime) {for (const enemy of Game.enemies) {enemy.terminate();}}

async function OnBeforeProjectStart(runtime) {runtime.addEventListener("tick", () => Tick(runtime));}

runOnStartup(async runtime => {
	runtime.addEventListener("beforeprojectstart", () => OnBeforeProjectStart(runtime));
	runtime.addEventListener("beforeanylayoutstart", () => initializeGameObjects(runtime));
	runtime.addEventListener("beforeanylayoutend", () => endLayout(runtime));
	runtime.addEventListener("keydown", e => {
		if (e.code == "ArrowUp" || e.code == "ArrowDown" || e.code == "KeyM") {Game.player.dash();}
		if (e.code == "KeyS") {Game.player.instance.behaviors.Platform.fallThrough();}
		if (e.code == "ArrowLeft" || e.code == "KeyN") {Game.player.attack();}
		if (e.code == "KeyW") {Game.player.jump();}
		if (e.code == "KeyR") {Game.player.hp = Game.vars.playerMaxHp;}
	});
	globalThis.newWave =function({eyeCount = 0, eyeLvl = 0, beetleCount = 0, beetleLvl = 0, frogCount = 0, frogLvl = 0, snakeCount = 0, snakeLvl = 0, crystalCount = 0, crystalLvl = 0, fairyCount = 0, fairyLvl = 0}={}){
		Game.enemies = [];
		let spawners = runtime.objects.Spawner.getAllInstances();
		function randomSpawn(enemy) {let index = Math.floor(Math.random() * spawners.length); let spawnLocation = spawners[index]; let curEnemy = enemy.createInstance(1, spawnLocation.x, spawnLocation.y); spawners.splice(index, 1);}

		for (let i = 0; i < eyeCount; i++) {randomSpawn(runtime.objects.FlyingEye);}																		for (let i = 0; i < beetleCount; i++) {randomSpawn(runtime.objects.Beetle);}																		for (let i = 0; i < frogCount; i++) {randomSpawn(runtime.objects.Frog);}																			for (let i = 0; i < snakeCount; i++) {randomSpawn(runtime.objects.Snake);}																			for (let i = 0; i < fairyCount; i++) {randomSpawn(runtime.objects.Fairy);}																			for (let i = 0; i < crystalCount; i++) {randomSpawn(runtime.objects.Crystal);}																		

		const eyes = runtime.objects.FlyingEye.getAllInstances(); const beetles = runtime.objects.Beetle.getAllInstances(); const frogs = runtime.objects.Frog.getAllInstances(); const snakes = runtime.objects.Snake.getAllInstances(); const crystals = runtime.objects.Crystal.getAllInstances(); const fairies = runtime.objects.Fairy.getAllInstances();	// Initialize enemy instances

		for (const instance of eyes) {Game.enemies.push(new LaserEnemy(instance, playerInstance, runtime, eyeLvl + runtime.globalVars.lvlBoost));}																																			for (const instance of beetles) {Game.enemies.push(new RushEnemy(instance, playerInstance, runtime, beetleLvl + runtime.globalVars.lvlBoost));}									for (const instance of frogs) {Game.enemies.push(new JumpEnemy(instance, playerInstance, runtime, frogLvl + runtime.globalVars.lvlBoost));}										for (const instance of snakes) {Game.enemies.push(new WalkEnemy(instance, playerInstance, runtime, snakeLvl + runtime.globalVars.lvlBoost));}										for (const instance of fairies) {Game.enemies.push(new FairyEnemy(instance, playerInstance, runtime, fairyLvl + runtime.globalVars.lvlBoost));}									for (const instance of crystals) {Game.enemies.push(new ProtectionEnemy(instance, playerInstance, runtime, crystalLvl + runtime.globalVars.lvlBoost));}
	}
});

function Tick(runtime) {
	if (!Game.player.instance) return;
	Game.player.update();										//player update
	Game.vars.playerHp = Game.player.hp;
	Game.enemies = Game.enemies.filter(enemy => enemy.instance !== null);
	Game.projectileDmg += (2 * runtime.globalVars.room)
	if ((Game.enemies.length == 0)) {Game.wave += 1; Game.nextWave = true;}		// waves
	if (runtime.layout.name == "Room1") {
		if (Game.wave == 1 && Game.nextWave == true) {newWave({frogCount: 1}); Game.nextWave = false;}														if (Game.wave == 2 && Game.nextWave == true) {newWave({frogCount: 2}); Game.nextWave = false;}														if (Game.wave == 3 && Game.nextWave == true) {newWave({snakeCount: 1, snakeLvl: 4}); Game.nextWave = false;}										if (Game.wave == 4 && Game.nextWave == true) {runtime.callFunction("nextLayout"); runtime.globalVars.room += 1;}}
	if (runtime.layout.name == "Room2") {
		if (Game.wave == 1 && Game.nextWave == true) {newWave({frogCount: 2, frogLvl: 2}); Game.nextWave = false;}											if (Game.wave == 2 && Game.nextWave == true) {newWave({eyeCount: 1, eyeLvl: 3}); Game.nextWave = false;}											if (Game.wave == 3 && Game.nextWave == true) {newWave({eyeCount: 2, eyeLvl: 1}); Game.nextWave = false;}											if (Game.wave == 4 && Game.nextWave == true) {newWave({eyeCount: 1, eyeLvl: 1, snakeCount: 1, snakeLvl: 1}); Game.nextWave = false;}				if (Game.wave == 4 && Game.nextWave == true) {newWave({eyeCount: 2, frogCount: 1, frogLvl: 4}); Game.nextWave = false;}								if (Game.wave == 5 && Game.nextWave == true) {newWave({eyeCount: 3, eyeLvl: 3}); Game.nextWave = false;}											if (Game.wave == 6 && Game.nextWave == true) {runtime.callFunction("nextLayout"); runtime.globalVars.room += 1;}}
	if (runtime.layout.name == "Room3") {
		if (Game.wave == 1 && Game.nextWave == true) {newWave({snakeCount: 3, snakeLvl: 1}); Game.nextWave = false;}										if (Game.wave == 2 && Game.nextWave == true) {newWave({snakeCount: 2, eyeCount: 1}); Game.nextWave = false;}										if (Game.wave == 3 && Game.nextWave == true) {newWave({beetleCount: 1, beetleLvl: 5}); Game.nextWave = false;}										if (Game.wave == 4 && Game.nextWave == true) {newWave({beetleCount: 3}); Game.nextWave = false;}													if (Game.wave == 5 && Game.nextWave == true) {newWave({beetleCount: 2, beetleLvl: 2, snakeCount: 2}); Game.nextWave = false;}						if (Game.wave == 6 && Game.nextWave == true) {newWave({beetleCount: 2, beetleLvl: 3, eyeCount: 2, eyeLvl: 1}); Game.nextWave = false;}				if (Game.wave == 7 && Game.nextWave == true) {newWave({frogCount: 4}); Game.nextWave = false;}														if (Game.wave == 8 && Game.nextWave == true) {runtime.callFunction("nextLayout"); runtime.globalVars.room += 1;}}
	if (runtime.layout.name == "Room4") {
		if (Game.wave == 1 && Game.nextWave == true) {newWave({fairyCount: 1}); Game.nextWave = false;}														if (Game.wave == 2 && Game.nextWave == true) {newWave({eyeCount: 2, eyeLvl: 4}); Game.nextWave = false;}											if (Game.wave == 3 && Game.nextWave == true) {newWave({eyeCount: 4}); Game.nextWave = false;}														if (Game.wave == 4 && Game.nextWave == true) {newWave({beetleCount: 2, beetleLvl: 3, frogCount: 2, frogLvl: 1}); Game.nextWave = false;}			if (Game.wave == 5 && Game.nextWave == true) {newWave({fairyCount: 2, fairyLvl: 2, beetleCount: 2, beetleLvl: 1}); Game.nextWave = false;}			if (Game.wave == 6 && Game.nextWave == true) {newWave({fairyCount: 3, eyeCount: 1, eyeLvl: 9}); Game.nextWave = false;}								if (Game.wave == 7 && Game.nextWave == true) {newWave({eyeCount: 2, eyeLvl: 5}); Game.nextWave = false;}											if (Game.wave == 8 && Game.nextWave == true) {newWave({beetleCount: 4}); Game.nextWave = false;}													if (Game.wave == 9 && Game.nextWave == true) {newWave({fairyCount: 3, eyeCount: 1, eyeLvl: 10}); Game.nextWave = false;}							if (Game.wave == 10 && Game.nextWave == true) {runtime.callFunction("nextLayout"); runtime.globalVars.room += 1;}}
	if (runtime.layout.name == "Room5") {
		if (Game.wave == 1 && Game.nextWave == true) {newWave({frogCount: 1, frogLvl: 5, crystalCount: 1}); Game.nextWave = false;}							if (Game.wave == 2 && Game.nextWave == true) {newWave({frogCount: 3, frogLvl: 1, crystalCount: 1, crystalLvl: 1}); Game.nextWave = false;}			if (Game.wave == 3 && Game.nextWave == true) {newWave({snakeCount: 2, snakeLvl: 5, beetleCount: 1, crystalCount: 1}); Game.nextWave = false;}		if (Game.wave == 4 && Game.nextWave == true) {newWave({eyeCount: 3, crystalCount: 1, crystalLvl: 5}); Game.nextWave = false;}						if (Game.wave == 5 && Game.nextWave == true) {newWave({crystalCount: 2, crystalLvl: 1, frogCount: 1,}); Game.nextWave = false;}						if (Game.wave == 6 && Game.nextWave == true) {runtime.callFunction("nextLayout"); runtime.globalVars.room += 1;}}					
	if (runtime.layout.name == "Room6") {
		if (Game.wave == 1 && Game.nextWave == true) {newWave({crystalCount: 3, beetleCount: 1, beetleLvl: 4}); Game.nextWave = false;}						if (Game.wave == 2 && Game.nextWave == true) {newWave({crystalCount: 2, eyeCount: 2}); Game.nextWave = false;}										if (Game.wave == 3 && Game.nextWave == true) {newWave({crystalCount: 2, fairyCount: 2, fairyLvl: 2}); Game.nextWave = false;}						if (Game.wave == 4 && Game.nextWave == true) {newWave({crystalCount: 3, eyeCount: 1, eyeLvl: 5}); Game.nextWave = false;}							if (Game.wave == 5 && Game.nextWave == true) {runtime.callFunction("nextLayout"); runtime.globalVars.room += 1;}}
	if (runtime.layout.name == "Room7") {
		if (Game.wave == 1 && Game.nextWave == true) {newWave({snakeCount: 3, snakeLvl: 7}); Game.nextWave = false;}										if (Game.wave == 2 && Game.nextWave == true) {newWave({snakeCount: 2, snakeLvl: 3, crystalCount: 1, crystalLvl: 3}); Game.nextWave = false;}		if (Game.wave == 3 && Game.nextWave == true) {newWave({snakeCount: 2, snakeLvl: 1, crystalCount: 2, crystalLvl: 2}); Game.nextWave = false;}		if (Game.wave == 4 && Game.nextWave == true) {newWave({beetleCount: 2, beetleLvl: 3, snakeCount: 2, snakeLvl: 2}); Game.nextWave = false;}			if (Game.wave == 5 && Game.nextWave == true) {newWave({eyeCount: 5, eyeLvl: 3}); Game.nextWave = false;}											if (Game.wave == 11 && Game.nextWave == true) {runtime.callFunction("nextLayout"); runtime.globalVars.room += 1;}}
	if (runtime.layout.name == "Room8") {
		if (Game.wave == 1 && Game.nextWave == true) {newWave({eyeCount: 4, eyeLvl: 2, crystalCount: 1, crystalLvl: 5}); Game.nextWave = false;}			if (Game.wave == 2 && Game.nextWave == true) {newWave({fairyCount: 4, fairyLvl: 4, crystalCount: 1, crystalLvl: 6}); Game.nextWave = false;}		if (Game.wave == 3 && Game.nextWave == true) {newWave({fairyCount: 3, snakeCount: 2, snakeLvl: 4}); Game.nextWave = false;}							if (Game.wave == 4 && Game.nextWave == true) {newWave({eyeCount: 4, eyeCount: 2, fairyCount: 1, fairyLvl: 6}); Game.nextWave = false;}				if (Game.wave == 5 && Game.nextWave == true) {newWave({snakeCount: 4, snakeLvl: 4, frogCount: 1, frogLvl: 7}); Game.nextWave = false;}				if (Game.wave == 6 && Game.nextWave == true) {runtime.callFunction("nextLayout"); runtime.globalVars.room += 1;}}
	if (runtime.layout.name == "Room9") {
		if (Game.wave == 1 && Game.nextWave == true) {newWave({}); Game.nextWave = false;}
		if (Game.wave == 2 && Game.nextWave == true) {runtime.callFunction("ending");}}
	


	const bullets = runtime.objects.Bullet.getAllInstances();	// bullet logic
	for (const bullet of bullets) {
		Game.player.collision(bullet, Game.vars.projectileDmg + runtime.globalVars.bulletBoost);
		if (bullet.testOverlap(Game.slashHitbox) && Game.player.isAttacking && Game.vars.redSword) {
			runtime.objects.BulletParticles.createInstance(1, bullet.x, bullet.y);	bullet.destroy();}
	}
	const lasers = runtime.objects.Laser.getAllInstances();		// laser logic
	for (const laser of lasers) {if(laser.instVars.damageActive) {Game.player.collision(laser, Game.vars.projectileDmg + runtime.globalVars.laserBoost);}}
	const bossLasers = runtime.objects.Laser2.getAllInstances();
	for (const laser of bossLasers) {if(laser.instVars.damageActive) {Game.player.collision(laser, 10 + runtime.globalVars.laserBoost);}}
	for (const enemy of Game.enemies) {	 						//enemy collision and AI updating
		Game.player.collision(enemy.instance, enemy.damage); enemy.update(); if (enemy.instance == null) return;
		enemy.instance.moveAdjacentToInstance(Game.player.instance, false); enemy.hpBar.moveAdjacentToInstance(Game.player.instance, false);
	}
	for (const spike of runtime.objects.Spike.getAllInstances()) {Game.player.collision(spike, 15);}	// Spike logic
	Game.hpPickups = runtime.objects.HPpickup.getAllInstances();	//health pickups
	for (const pickup of Game.hpPickups) {
		if (pickup.testOverlap(Game.player.instance)) {pickup.destroy(); Game.player.hp += 10;}
	}
}
function randInt(min, max) {return Math.floor(Math.random() * (max - min + 1)) + min;}