export class PlayerClass {
    constructor(instance, hp, runtime) {
        this.instance = instance; this.platform = instance.behaviors.Platform; this.runtime = runtime;
        this.hp = hp; this.maxHp = Game.vars.playerMaxHp; this.isAttacking = false; this.attackCooldown = 0; this.canJump = true;
        this.jumpCount = 0;
        this.knockbackVelocity = { x: 0, y: 0};
        this.knockbackDecay = 0.7;  // How fast knockback slows (0.9 = 10% loss per tick)
        this.minKnockbackSpeed = 1; // Stop knockback when speed is low
        this.invulnerableTime = 0;  // i-frame duration in seconds
        this.dashDuration = 0; this.dashCooldown = 0; this.dashBar = null;
        this.hpBar = this.runtime.objects.HpBar.createInstance(1, 5, 12); this.hpFill = this.runtime.objects.HpFill.createInstance(1, 5, 12);
    }
    update() {
        this.controller(); this.animator(); this.updateKnockback(); this.platform.jumpStrength = Game.vars.jumpStrength;
        Game.damageFlash.moveToTop(); this.instance.moveAdjacentToInstance(Game.damageFlash, false); this.platform.maxSpeed = 125;

        Game.slashHitbox.x = playerInstance.x; Game.slashHitbox.y = playerInstance.y; Game.slashHitbox.width = playerInstance.width;
        Game.damageFlash.x = playerInstance.x; Game.damageFlash.y = playerInstance.y;

        this.hpFill.width = 133 * (this.hp / this.maxHp);   // health logic
        if (this.hp > this.maxHp) {this.hp = this.maxHp;}

        if (this.invulnerableTime > 0) {this.invulnerableTime -= this.runtime.dt; Game.damageFlash.isVisible = true;}
        else {Game.damageFlash.isVisible = false;}

        if (this.hp <= 0) {this.destroy();}

        if (this.platform.isFalling && (this.platform.isByWall("left") || (this.platform.isByWall("right")))) {  // Wall Slide
            if (this.platform.isByWall("left") && this.runtime.keyboard.isKeyDown("KeyA")) {
                this.platform.maxFallSpeed = 50; this.instance.width = 24;
            }
            if (this.platform.isByWall("right") && this.runtime.keyboard.isKeyDown("KeyD")) {
                this.platform.maxFallSpeed = 50; this.instance.width = -24;
            }
        } else {this.platform.maxFallSpeed = Game.vars.fallSpeed;}
        if (this.dashDuration > 0) {    // Dash duration 
            this.dashDuration -= this.runtime.dt;
            this.platform.vectorY = 0;
            let dashEffect = this.runtime.objects.PlayerAfterImage.createInstance(1, this.instance.x, this.instance.y); dashEffect.moveToBottom();
        }
        if (this.dashCooldown > 0) {    // Dash cooldown
            this.dashCooldown -= this.runtime.dt;
            this.dashBar.x = this.instance.x; this.dashBar.y = this.instance.y; this.dashBar.width = this.dashCooldown * 20;
        }
        if (this.attackCooldown > 0) {this.attackCooldown -= this.runtime.dt;}
        else {this.isAttacking = false;}
    }
    controller() {
        const keyboard = this.runtime.keyboard;
        if (this.platform.isOnFloor || (this.platform.isByWall("left") || this.platform.isByWall("right"))) {this.jumpCount = 0;}

        if (keyboard.isKeyDown("KeyA")) {this.platform.simulateControl("left"); this.instance.width = -24;}
        if (keyboard.isKeyDown("KeyD")) {this.platform.simulateControl("right"); this.instance.width = 24;}

        this.canJump = this.platform.isOnFloor || (((this.platform.isByWall("left") || this.platform.isByWall("right")) && this.platform.isFalling) || this.jumpCount < Game.vars.maxJumps);
    }
    jump() {
        if (!this.canJump) return;
        this.platform.vectorY -= this.platform.jumpStrength; this.runtime.callFunction("playAudio", "Jump", -15, false, ""); this.jumpCount += 1;
        if (this.platform.isByWall("left")) {this.applyKnockback(this.instance.x - 1, this.instance.y, 400);}
        if (this.platform.isByWall("right")) {this.applyKnockback(this.instance.x + 1, this.instance.y, 400);}
    }
    dash() {
        if (this.dashCooldown > 0) return;
        this.dashCooldown = Game.vars.dashCooldown; this.dashDuration = Game.vars.dashStrength/(Game.vars.dashStrength * 4);
        if (this.instance.width == 24) {this.applyKnockback(this.instance.x - 1, this.instance.y, Game.vars.dashStrength);}
        else if (this.instance.width == -24) {this.applyKnockback(this.instance.x + 1, this.instance.y, Game.vars.dashStrength)}
        this.dashBar = this.runtime.objects.DashCooldown.createInstance(1, this.instance.x, this.instance.y);
        this.runtime.callFunction("PlayAudio", "Dash3", -15, false, "");
    }
    attack() {
        if (this.attackCooldown > 0) return;
        this.isAttacking = true;
        this.runtime.callFunction("PlayAudio", "Cut", -10, false, "");
        if (Game.vars.holySword) {
            let swordBeam = this.runtime.objects.SwordBeam.createInstance(1, this.instance.x, this.instance.y);
            if (this.instance.width == 24) {swordBeam.angle = 0;} else {swordBeam.angle = Math.PI;}
        }
        this.attackCooldown = .15;
    }
    animator() {
        if (this.instance == null) return;
        const keyboard = this.runtime.keyboard; const playerState = this.platform;
        if (this.isAttacking) {this.instance.setAnimation("Slash");}
        else if (playerState.isJumping) {this.instance.setAnimation("Jump");}
        else if (playerState.isFalling) {this.instance.setAnimation("WallSlide");}
        else if ((playerState.isMoving) && (playerState.isOnFloor)) {this.instance.setAnimation("Walk");}
        else if (playerState.isOnFloor) {this.instance.setAnimation("Idle");}
    }
    collision(enemyInstance, damage) {
        if (this.instance.testOverlap(enemyInstance)) {this.damage(damage);}
    }
    applyKnockback(sourceX, sourceY, magnitude) {
        const dx = this.instance.x - sourceX; const dy = this.instance.y - sourceY; const dist = Math.sqrt(dx * dx + dy * dy);
        if (dist === 0) {return};
        const nx = dx / dist; const ny = dy / dist;
	    this.knockbackVelocity.x += nx * magnitude; this.knockbackVelocity.y += ny * magnitude;
    }
    updateKnockback() {
		const v = this.knockbackVelocity; const speed = Math.sqrt(v.x * v.x + v.y * v.y);
		if (speed > this.minKnockbackSpeed) {
			this.instance.x += v.x * this.runtime.dt; this.instance.y += v.y * this.runtime.dt;
            v.x *= this.knockbackDecay; v.y *= this.knockbackDecay;
		} else {v.x = 0; v.y = 0;}
	}
    damage(damage) {
        if (this.invulnerableTime > 0 || (Game.vars.iframeDash && this.dashDuration > 0)) return; 
        let dmgTaken = (damage - Game.vars.playerDef); if (dmgTaken <= 0) {dmgTaken = 1;}
        this.hp -= dmgTaken; console.log("damage: ", dmgTaken); this.invulnerableTime = Game.vars.playerInvulnerabilty; this.runtime.globalVars.ScreenShake = true;
        this.runtime.callFunction("PlayAudio", "HurtSound", 0, false, "");
        const damageParticles = this.runtime.objects.PlayerDamageParticles.createInstance(1, this.instance.x, this.instance.y);
    }  
    destroy() {
        if (Game.vars.alive == true) {
            this.instance.x = 640; this.instance.y = 360; this.runtime.callFunction("PlayAudio", "DeathSound", -10, false, "");
            this.runtime.callFunction("death"); Game.vars.alive = false;
            for (const enemy of Game.enemies) {enemy.terminate();}
        }
    }
}